/*
 *   Copyright (C) 2009-2015 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Config.h"
#include "Globals.h"
#include "YSFRX.h"
#include "Utils.h"

const unsigned int BUFFER_LENGTH = 200U;

const q15_t SCALING_FACTOR1 = 16384;      // Q15(0.5)        // XXX
const q15_t SCALING_FACTOR2 = 1365;       // Q15(1 / 24)

const uint32_t PLLMAX = 0x10000U;
const uint32_t PLLINC = PLLMAX / YSF_RADIO_SYMBOL_LENGTH;
const uint32_t INC    = PLLINC / 32U;

const uint8_t SYNC_SYMBOL_ERRS = 2U;
const uint8_t SYNC_BIT_ERRS    = 3U;

const unsigned int MAX_SYNC_FRAMES = 4U;

const uint8_t BIT_MASK_TABLE[] = {0x80U, 0x40U, 0x20U, 0x10U, 0x08U, 0x04U, 0x02U, 0x01U};

#define WRITE_BIT(p,i,b) p[(i)/8] = (b) ? (p[(i)/8] | BIT_MASK_TABLE[(i)%8]) : (p[(i)/8] & ~BIT_MASK_TABLE[(i)%8])
#define READ_BIT(p,i)    (p[(i)/8] & BIT_MASK_TABLE[(i)%8])

CYSFRX::CYSFRX() :
m_pll(0U),
m_prev(false),
m_state(YSFRXS_NONE),
m_bitBuffer(0x00U),
m_symbols(),
m_buffer(),
m_bufferPtr(0U),
m_symbolPtr(0U),
m_lostCount(0U),
m_centre(0),
m_threshold(0)
{
}

void CYSFRX::samples(const q15_t* samples, uint8_t length)
{
  for (uint16_t i = 0U; i < length; i++) {
    bool bit = samples[i] < 0;

    if (bit != m_prev) {
      if (m_pll < (PLLMAX / 2U))
        m_pll += INC;
      else
        m_pll -= INC;
    }

    m_prev = bit;

    m_pll += PLLINC;

    if (m_pll >= PLLMAX) {
      m_pll -= PLLMAX;

      if (m_state == YSFRXS_NONE)
        processNone(samples[i]);
      else
        processData(samples[i]);
    }
  }
}

void CYSFRX::processNone(q15_t sample)
{
  m_symbolBuffer <<= 1;
  if (sample < 0)
    m_symbolBuffer |= 0x01U;

  m_symbols[m_symbolPtr] = sample;

  // Fuzzy matching of the data sync bit sequence
  if (countBits32((m_symbolBuffer & YSF_SYNC_SYMBOLS_MASK) ^ YSF_SYNC_SYMBOLS) <= SYNC_SYMBOL_ERRS) {
    uint16_t ptr = (m_symbolPtr + 1U) % YSF_SYNC_LENGTH_SYMBOLS;

    q31_t mean = 0;
    q15_t max = -16000;

    for (uint8_t i = 0U; i < YSF_SYNC_LENGTH_SYMBOLS; i++) {
      if (m_symbols[ptr] > max)
        max = m_symbols[ptr];

      mean += m_symbols[ptr];

      ptr = (ptr + 1U) % YSF_SYNC_LENGTH_SYMBOLS;
    }

    q31_t v2 = mean * SCALING_FACTOR2;
    q15_t centre = q15_t(__SSAT(v2 >> 15, 16));

    q31_t v1 = (max - centre) * SCALING_FACTOR1;
    q15_t threshold = q15_t(v1 >> 15);

    uint8_t sync[YSF_SYNC_LENGTH_BYTES];
    uint16_t offset = 0U;

    ptr = (m_symbolPtr + 1U) % YSF_SYNC_LENGTH_SYMBOLS;
    for (uint8_t i = 0U; i < YSF_SYNC_LENGTH_SYMBOLS; i++) {
      q15_t sample = m_symbols[ptr] - centre;

      if (sample < -threshold) {
        m_bitBuffer <<= 2;
        m_bitBuffer |= 0x01U;
      } else if (sample < 0) {
        m_bitBuffer <<= 2;
        m_bitBuffer |= 0x00U;
      } else if (sample < threshold) {
        m_bitBuffer <<= 2;
        m_bitBuffer |= 0x02U;
      } else {
        m_bitBuffer <<= 2;
        m_bitBuffer |= 0x03U;
      }

      ptr = (ptr + 1U) % YSF_SYNC_LENGTH_SYMBOLS;
    }

    // Fuzzy matching of the data sync bit sequence
    if (countBits64((m_bitBuffer & YSF_SYNC_BITS_MASK) ^ YSF_SYNC_BITS) <= SYNC_BIT_ERRS) {
      DEBUG3("YSFRX: Sync found in None centre/threshold", centre, threshold);
      for (uint8_t i = 0U; i < YSF_SYNC_LENGTH_BYTES; i++)
        m_buffer[i] = YSF_SYNC_BYTES[i];
      m_centre    = centre;
      m_threshold = threshold;
      m_lostCount = MAX_SYNC_FRAMES;
      m_bufferPtr = YSF_SYNC_LENGTH_BITS;
      m_state     = YSFRXS_DATA;
      io.setDecode(true);
    }
  }

  m_symbolPtr = (m_symbolPtr + 1U) % YSF_SYNC_LENGTH_SYMBOLS;
}

void CYSFRX::processData(q15_t sample)
{
  sample -= m_centre;

  if (sample < -m_threshold) {
    m_bitBuffer <<= 2;
    m_bitBuffer |= 0x01U;
    WRITE_BIT(m_buffer, m_bufferPtr, false);
    m_bufferPtr++;
    WRITE_BIT(m_buffer, m_bufferPtr, true);
    m_bufferPtr++;
  } else if (sample < 0) {
    m_bitBuffer <<= 2;
    m_bitBuffer |= 0x00U;
    WRITE_BIT(m_buffer, m_bufferPtr, false);
    m_bufferPtr++;
    WRITE_BIT(m_buffer, m_bufferPtr, false);
    m_bufferPtr++;
  } else if (sample < m_threshold) {
    m_bitBuffer <<= 2;
    m_bitBuffer |= 0x02U;
    WRITE_BIT(m_buffer, m_bufferPtr, true);
    m_bufferPtr++;
    WRITE_BIT(m_buffer, m_bufferPtr, false);
    m_bufferPtr++;
  } else {
    m_bitBuffer <<= 2;
    m_bitBuffer |= 0x03U;
    WRITE_BIT(m_buffer, m_bufferPtr, true);
    m_bufferPtr++;
    WRITE_BIT(m_buffer, m_bufferPtr, true);
    m_bufferPtr++;
  }

  // Only search for a sync in the right place +-1 symbol
  if (m_bufferPtr >= (YSF_SYNC_LENGTH_BITS - 2U) && m_bufferPtr <= (YSF_SYNC_LENGTH_BITS + 2U)) {
    // Fuzzy matching of the data sync bit sequence
    if (countBits64((m_bitBuffer & YSF_SYNC_BITS_MASK) ^ YSF_SYNC_BITS) <= SYNC_BIT_ERRS) {
#if defined(WANT_DEBUG)
      if (m_bufferPtr < YSF_SYNC_LENGTH_BITS)
        DEBUG2("YSFRX: Found sync in Data, early", YSF_SYNC_LENGTH_BITS - m_bufferPtr);
      else if (m_bufferPtr > YSF_SYNC_LENGTH_BITS)
        DEBUG2("YSFRX: Found sync in Data, late", m_bufferPtr - YSF_SYNC_LENGTH_BITS);
      else
        DEBUG1("YSFRX: Found sync in Data");
#endif
      m_lostCount = MAX_SYNC_FRAMES;
      m_bufferPtr = YSF_SYNC_LENGTH_BITS;
    }
  }

  // Send a data frame to the host if the required number of bits have been received, or if a data sync has been seen
  if (m_bufferPtr == YSF_FRAME_LENGTH_BITS) {
    // We've not seen a data sync for too long, signal RXLOST and change to RX_NONE
    m_lostCount--;
    if (m_lostCount == 0U) {
      DEBUG1("YSFRX: Sync timed out, lost lock");
      io.setDecode(false);

      serial.writeYSFLost();

      m_state = YSFRXS_NONE;
      return;
    }

    // XXX Check the FICH for EOT

    serial.writeYSFData(m_buffer, YSF_FRAME_LENGTH_BYTES);

    // Start the next frame
    ::memset(m_buffer, 0x00U, YSF_FRAME_LENGTH_BYTES);
    m_bufferPtr = 0U;
  }
}

uint8_t CYSFRX::writeEOT()
{
  if (m_state == YSFRXS_DATA)
    return 4U;

  m_state = YSFRXS_NONE;
  io.setDecode(false);

  return 0U;
}

